#define _WINSOCK_DEPRECATED_NO_WARNINGS
#pragma comment(lib, "Ws2_32.lib")

#include "TCPModbusIntf.h"

#include <tchar.h>

TCPModbusIntf* tcpModbus;

using namespace std;

void Usage() {
	cout << "================================================================================\n";
	cout << "                              ModbusTcp Client App Usage\n\n";
	cout << " MBusTcpClientApp -caddr [Client(Local) IP Address] -saddr [Server IP Address]\n";
	cout << "                  -cport[Client Port] -cnt[Loop Count]\n";
	cout << "                  -hregnum [Total holding register number]\n";
	cout << "                  -iregnum [Total input register number]\n\n";
	cout << " Example:\n";
	cout << " MBusTcpClientApp -caddr 172.19.80.1 -saddr 172.19.80.2 -cport 7000 -cnt 10 -hregnum 8 -iregnum 16\n";
	cout << "================================================================================\n";

	char key;
	cout << "\nPlease type any key to continue (q for exit)....";
	cin >> key;
	if (key == 'q' || key == 'Q')
		exit(EXIT_SUCCESS);
	cout << endl;
}

///////////////////////////////////////////////////////////////////////////////////////////////
/// main()
///////////////////////////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
	const char szFunc[] = "main()";

	std::string szLocalIpAddr = "172.19.80.1";     // Mobus/TCP Client IP Address
	std::string szServerIpAddr = "172.19.80.1";    // Mobus/TCP Server IP Address
	WORD wLocalPort = 7000;   // Mobus/TCP Client port
	WORD wServerPort = 502;   // Mobus/TCP Server port
	int loopCount = 1;
	int maxHoldRegNum = 1;
	int maxInputRegNum = 1;

	Usage();

	// process parameters
	int i = 1;
	while (i < argc) {
		std::string param(argv[i]);
		if (param.compare("-caddr") == 0) {
			szLocalIpAddr = std::string(argv[++i]);
		} 
		else if (param.compare("-saddr") == 0) {
			szServerIpAddr = argv[++i];
		}
		else if (param.compare("-cport") == 0) {
			wLocalPort = atoi(argv[++i]);
		}
		else if (param.compare("-cnt") == 0) {
			loopCount = atoi(argv[++i]);
		}
		else if (param.compare("-hregnum") == 0) {
			maxHoldRegNum = atoi(argv[++i]);
		}
		else if (param.compare("-iregnum") == 0) {
			maxInputRegNum = atoi(argv[++i]);
		}
		else {
			cout << "Invalid input parameters. Exit\n";
			exit(EXIT_FAILURE);
		}
		i++;
	}

	printf("Your input is:\nClientIpAddr=%s, ServerIpAddr=%s, ClientPort=%d\nLoopCount=%d, MaxHoldingRegistersNumber=%d, MaxInputRegisterNum=%d\n\n",
		szLocalIpAddr, szServerIpAddr, wLocalPort, loopCount, maxHoldRegNum, maxInputRegNum);

	tcpModbus = TCPModbusIntf::GetInstance();
	if (tcpModbus == NULL) {
		printf("Failed to new ModBusTcp object. Exit!!!\n");
		exit(EXIT_FAILURE);
	}

	BOOL bSuccess = TRUE;
    std::cout << "=== Modbus/TCP Client ===\n";

	WSADATA wsa;
	if (WSAStartup(MAKEWORD(2, 2), &wsa) != 0)
	{
		printf("Failed. Error Code : %d\n", WSAGetLastError());
		exit(EXIT_FAILURE);
	}

	if (!tcpModbus->Initialize(0, szLocalIpAddr, wLocalPort, szServerIpAddr))
	{
		bSuccess = FALSE;
		printf("%s: LocalIpAddr=%s, ServerIPAddr=%s, LocalPort=%d\n", 
			szFunc, szLocalIpAddr.c_str(), szServerIpAddr.c_str(), wLocalPort);
		exit(EXIT_FAILURE);
	}
	else
	{
		printf("%s: Succeed to initialize the Modbus/TCP\n", szFunc);
	}

	// Parameters
	WORD wChannelNo = 5;
	BYTE ucOutValue[2];
	ucOutValue[0] = 0xaa;  ucOutValue[1] = 0xaa;
	BYTE ucInValue[2];
	ucInValue[0] = 0;      ucInValue[1] = 0;

	const int NumOfRegs = 16;
	WORD wOutValue[NumOfRegs];  // 16 register values: 0xfff0, 0xfff1, 0xfff2,.. 0xffff,
	for (int i = 0; i < NumOfRegs; i++) {
		wOutValue[i] = (0xff << 8) + (0xff & i);
	}
	WORD wInValue[NumOfRegs];  // reset to 0x0000
	for (int i = 0; i < NumOfRegs; i++) {
		wInValue[i] = 0;
	}


	bool isStopped = false;
	int testCount = 0;
	int testNo = 1;
	while (! isStopped) {
		testCount++;
		printf("\n============================ TEST %d ============================\n", testCount);

		for (int i = 0; i < loopCount; i++) {

			// Test all Modbus Functions 
			//tcpModbus.TestMBusTCP();

			{ // MbPresetSingleReg test - 4XXXX holding register
				WORD wChannelNo = 5;

				printf("\n4XXXX Test %d: MbPresetSingleReg (channelNo=%d, Value=%02x)\n", testNo++, wChannelNo, wOutValue[wChannelNo - 1]);

				if (FALSE == tcpModbus->WriteSingleHoldReg(wChannelNo, wOutValue[wChannelNo - 1])) {
					printf("MbPresetSingleReg KO. Error: %s\n", tcpModbus->LastErrorString());
				}
				else printf("MbPresetSingleReg OK!\n");
			}

			{ // MbReadHoldingRegs test - 4XXXX holding register
				WORD wChannelNo = 5;

				printf("\n4XXXX Test %d: MbReadHoldingRegs (channelNo=%d)\n", testNo++, wChannelNo);

				if (FALSE == tcpModbus->ReadHoldRegs(wChannelNo, &wInValue[wChannelNo - 1], 1)) {
					printf("MbReadHoldingRegs KO. Error: %s\n", tcpModbus->LastErrorString());
				}
				else printf("MbReadHoldingRegs OK!\n");

				if (wOutValue[wChannelNo - 1] != wInValue[wChannelNo - 1])
					printf("KO! Write & Read data are NOT equal!\n");
				else
					printf("OK! Write & Read data are equal!\n");
			}

			{ // WriteMultiHoldRegs test - 4XXXX holding register
				WORD wChannelNo = 0;
				WORD wChannelCount = maxHoldRegNum;

				printf("\n4XXXX Test %d: WriteMultiHoldRegs (channelNo=%d, channelCount=%d)\n", testNo++, wChannelNo, wChannelCount);

				if (FALSE == tcpModbus->WriteMultiHoldRegs(wChannelNo, wOutValue, wChannelCount)) {
					printf("WriteMultiHoldRegs KO. Error: %s\n", tcpModbus->LastErrorString());
				}
				else printf("WriteMultiHoldRegs OK!\n");
			}

			{ // ReadHoldRegs test - 4XXXX holding register
				WORD wChannelNo = 0;
				WORD wChannelCount = maxHoldRegNum;

				printf("\n4XXXX Test %d: ReadHoldRegs (channelNo=%d, channelCount=%d)\n", testNo++, wChannelNo, wChannelCount);

				if (FALSE == tcpModbus->ReadHoldRegs(wChannelNo, wInValue, wChannelCount)) {
					printf("ReadHoldRegs KO. Error: %s\n", tcpModbus->LastErrorString());
				}
				else printf("ReadHoldRegs OK!\n");

				bool bEqual = true;
				for (int i = 0; i < wChannelCount; i++) {
					if (wOutValue[i] != wInValue[i])   bEqual = false;
				}
				if (!bEqual)
					printf("KO! Write & Read data are NOT equal!\n");
				else
					printf("OK! Write & Read data are equal!\n");
			}

			{ // ReadInputRegs test - 3XXXX input register
				WORD wChannelNo = 0;
				WORD wChannelCount = maxInputRegNum;

				printf("\n4XXXX Test %d: ReadInputRegs (channelNo=%d, channelCount=%d)\n", testNo++, wChannelNo, wChannelCount);

				if (FALSE == tcpModbus->ReadInputRegs(wChannelNo, wInValue, wChannelCount)) {
					printf("ReadInputRegs KO. Error: %s\n", tcpModbus->LastErrorString());
				}
				else printf("ReadInputRegs OK!\n");
			}

		}  // end of LoopCount

		// User interface - select Stop or go-on
		char input;
		cout << "\n--------\nSTOP the application? (y/n)" << endl;
		cin >> input;
		if (input == 'Y' || input == 'y')  isStopped = true;
	}

	// cleanup and exit
	tcpModbus->Close();
	exit(EXIT_SUCCESS);

}